// Copyright (C) 2024 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef PERFMON_LOG_H
#define PERFMON_LOG_H

#include <log/logger_support.h>
#include <log/macros.h>
#include <perfmon_messages.h>

namespace isc {
namespace perfmon {

extern isc::log::Logger perfmon_logger;

} // end of isc::perfmon namespace
} // end of isc namespace


#endif
